@@ -7,7 +7,7 @@ require_relative 'tmp_path'
 
 # Only single mode tests go here. Cluster and pumactl tests
 # have their own files, use those instead
-class TestIntegration < TimeoutTestCase
+class TestIntegration < PumaTest
   include TmpPath
   HOST  = "127.0.0.1"
   TOKEN = "xxyyzz"
@@ -515,6 +515,8 @@ class TestIntegration < TimeoutTestCase
             sleep 0.15 * (2 ** restart_count)
           rescue Minitest::Assertion # Timeout
             run = false
+          rescue Errno::EBADF # bad restart?
+            run = false
           end
         else
           sleep 0.1
