@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 require_relative "helper"
 require_relative "helpers/integration"
 
@@ -111,18 +113,14 @@ class TestIntegrationCluster < TestIntegration
   # No more than 10 should throw Errno::ECONNRESET.
 
   def test_term_closes_listeners_tcp
-    skip_unless_signal_exist? :TERM
     term_closes_listeners unix: false
   end
 
   def test_term_closes_listeners_unix
-    skip_unless_signal_exist? :TERM
     term_closes_listeners unix: true
   end
 
   def test_term_exit_code
-    skip_unless_signal_exist? :TERM
-
     cli_server "-w #{workers} test/rackup/hello.ru"
     _, status = stop_server
 
@@ -130,8 +128,6 @@ class TestIntegrationCluster < TestIntegration
   end
 
   def test_term_suppress
-    skip_unless_signal_exist? :TERM
-
     cli_server "-w #{workers} -C test/config/suppress_exception.rb test/rackup/hello.ru"
 
     _, status = stop_server
@@ -140,7 +136,6 @@ class TestIntegrationCluster < TestIntegration
   end
 
   def test_on_booted_and_on_stopped
-    skip_unless_signal_exist? :TERM
     cli_server "-w #{workers} test/rackup/hello.ru", config: <<~CONFIG
       on_booted  { STDOUT.syswrite "on_booted called\n"  }
       on_stopped { STDOUT.syswrite "on_stopped called\n" }
@@ -156,8 +151,20 @@ class TestIntegrationCluster < TestIntegration
     wait_server 15
   end
 
+  def test_on_booted_with_fork_worker_refork
+    cli_server "-w #{workers} test/rackup/hello.ru", config: <<~CONFIG
+      fork_worker
+      on_booted { STDOUT.syswrite "on_booted called\n" }
+    CONFIG
+
+    assert wait_for_server_to_include("on_booted called")
+
+    Process.kill :SIGURG, @pid
+
+    assert wait_for_server_to_include("on_booted called")
+  end
+
   def test_term_worker_clean_exit
-    skip_unless_signal_exist? :TERM
     cli_server "-w #{workers} test/rackup/hello.ru"
 
     # Get the PIDs of the child workers.
@@ -174,8 +181,6 @@ class TestIntegrationCluster < TestIntegration
 
   # mimicking stuck workers, test respawn with external TERM
   def test_stuck_external_term_spawn
-    skip_unless_signal_exist? :TERM
-
     worker_respawn(0) do |phase0_worker_pids|
       last = phase0_worker_pids.last
       # test is tricky if only one worker is TERM'd, so kill all but
@@ -193,8 +198,6 @@ class TestIntegrationCluster < TestIntegration
   # `Process.wait2(<child pid>)` still works properly. This bug has
   # been fixed in Ruby 3.3.
   def test_workers_respawn_with_process_detach
-    skip_unless_signal_exist? :KILL
-
     config = 'test/config/process_detach_before_fork.rb'
 
     worker_respawn(0, workers, config) do |phase0_worker_pids|
@@ -217,10 +220,28 @@ class TestIntegrationCluster < TestIntegration
 
   # mimicking stuck workers, test restart
   def test_stuck_phased_restart
-    skip_unless_signal_exist? :USR1
     worker_respawn { |phase0_worker_pids| Process.kill :USR1, @pid }
   end
 
+  def test_puma_stats_clustered
+    worker_check_interval = 2
+
+    cli_server "-w 1 -t 1:1 #{set_pumactl_args unix: true} test/rackup/hello.ru",
+      config: "worker_check_interval #{worker_check_interval}"
+
+    sleep worker_check_interval + 1
+
+    stats = get_stats
+
+    assert_instance_of Hash, stats
+
+    last_status = stats['worker_status'].first['last_status']
+
+    Puma::Server::STAT_METHODS.each do |stat|
+      assert_includes last_status, stat.to_s
+    end
+  end
+
   def test_worker_check_interval
     # iso8601 2022-12-14T00:05:49Z
     re_8601 = /\A\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z\z/
@@ -280,8 +301,6 @@ class TestIntegrationCluster < TestIntegration
   end
 
   def test_worker_index_is_with_in_options_limit
-    skip_unless_signal_exist? :TERM
-
     cli_server "-C test/config/t3_conf.rb test/rackup/hello.ru"
 
     get_worker_pids(0, 3) # this will wait till all the processes are up
@@ -549,8 +568,6 @@ class TestIntegrationCluster < TestIntegration
   end
 
   def test_hook_data
-    skip_unless_signal_exist? :TERM
-
     cli_server "-C test/config/hook_data.rb test/rackup/hello.ru"
     get_worker_pids 0, 2 # make sure workers are booted
     stop_server
@@ -611,7 +628,7 @@ class TestIntegrationCluster < TestIntegration
     pids = []
     re = /Terminating timed out worker \(Worker \d+ #{details}\): (\d+)/
 
-    Timeout.timeout(iterations * (timeout + 1)) do
+    Timeout.timeout(iterations * (timeout + 1.5)) do
       while (pids.size < workers * iterations)
         idx = wait_for_server_to_match(re, 1).to_i
         pids << idx
@@ -624,8 +641,6 @@ class TestIntegrationCluster < TestIntegration
   # Send requests 10 per second.  Send 10, then :TERM server, then send another 30.
   # No more than 10 should throw Errno::ECONNRESET.
   def term_closes_listeners(unix: false)
-    skip_unless_signal_exist? :TERM
-
     cli_server "-w #{workers} -t 0:6 -q test/rackup/sleep_step.ru", unix: unix
     threads = []
     replies = []
