@@ -1,17 +1,26 @@
+# frozen_string_literal: true
+
 require_relative "helper"
 require_relative "helpers/integration"
 
 class TestIntegrationSingle < TestIntegration
-  parallelize_me! if ::Puma.mri?
+  parallelize_me! if ::Puma::IS_MRI
 
   def workers ; 0 ; end
 
   def test_hot_restart_does_not_drop_connections_threads
-    hot_restart_does_not_drop_connections num_threads: 5, total_requests: 1_000
+    ttl_reqs = Puma.windows? ? 500 : 1_000
+    restart_does_not_drop_connections num_threads: 5, total_requests: ttl_reqs,
+      signal: :USR2
   end
 
   def test_hot_restart_does_not_drop_connections
-    hot_restart_does_not_drop_connections
+    if Puma.windows?
+      restart_does_not_drop_connections total_requests: 300,
+        signal: :USR2
+    else
+      restart_does_not_drop_connections signal: :USR2
+    end
   end
 
   def test_usr2_restart
@@ -36,14 +45,31 @@ class TestIntegrationSingle < TestIntegration
 
   def test_term_exit_code
     skip_unless_signal_exist? :TERM
-    skip_if :jruby # JVM does not return correct exit code for TERM
 
     cli_server "test/rackup/hello.ru"
     _, status = stop_server
 
+    status = status.exitstatus % 128 if ::Puma::IS_JRUBY
     assert_equal 15, status
   end
 
+  def test_on_booted_and_on_stopped
+    skip_unless_signal_exist? :TERM
+
+    cli_server "test/rackup/hello.ru", no_wait: true, config: <<~CONFIG
+      on_booted  { STDOUT.syswrite "on_booted called\n"  }
+      on_stopped { STDOUT.syswrite "on_stopped called\n" }
+    CONFIG
+
+    assert wait_for_server_to_include('on_booted called')
+
+    Process.kill :TERM, @pid
+
+    assert wait_for_server_to_include('on_stopped called')
+
+    wait_server 15
+  end
+
   def test_term_suppress
     skip_unless_signal_exist? :TERM
 
@@ -53,6 +79,28 @@ class TestIntegrationSingle < TestIntegration
     assert_equal 0, status
   end
 
+  def test_rack_url_scheme_default
+    skip_unless_signal_exist? :TERM
+
+    cli_server("test/rackup/url_scheme.ru")
+
+    reply = read_body(connect)
+    stop_server
+
+    assert_match("http", reply)
+  end
+
+  def test_conf_is_loaded_before_passing_it_to_binder
+    skip_unless_signal_exist? :TERM
+
+    cli_server("-C test/config/rack_url_scheme.rb test/rackup/url_scheme.ru")
+
+    reply = read_body(connect)
+    stop_server
+
+    assert_match("https", reply)
+  end
+
   def test_prefer_rackup_file_specified_by_cli
     skip_unless_signal_exist? :TERM
 
@@ -65,7 +113,6 @@ class TestIntegrationSingle < TestIntegration
 
   def test_term_not_accepts_new_connections
     skip_unless_signal_exist? :TERM
-    skip_if :jruby
 
     cli_server 'test/rackup/sleep.ru'
 
@@ -73,7 +120,7 @@ class TestIntegrationSingle < TestIntegration
     sleep 1 # ensure curl send a request
 
     Process.kill :TERM, @pid
-    true while @server.gets !~ /Gracefully stopping/ # wait for server to begin graceful shutdown
+    assert wait_for_server_to_include('Gracefully stopping') # wait for server to begin graceful shutdown
 
     # Invoke a request which must be rejected
     _stdin, _stdout, rejected_curl_stderr, rejected_curl_wait_thread = Open3.popen3("curl #{HOST}:#{@tcp_port}")
@@ -85,16 +132,13 @@ class TestIntegrationSingle < TestIntegration
     rejected_curl_wait_thread.join
 
     assert_match(/Slept 10/, curl_stdout.read)
-    assert_match(/Connection refused/, rejected_curl_stderr.read)
+    assert_match(/Connection refused|(Couldn't|Could not) connect to server/, rejected_curl_stderr.read)
 
-    Process.wait(@server.pid)
-    @server.close unless @server.closed?
-    @server = nil # prevent `#teardown` from killing already killed server
+    wait_server 15
   end
 
   def test_int_refuse
     skip_unless_signal_exist? :INT
-    skip_if :jruby  # seems to intermittently lockup JRuby CI
 
     cli_server 'test/rackup/hello.ru'
     begin
@@ -105,7 +149,7 @@ class TestIntegrationSingle < TestIntegration
     end
 
     Process.kill :INT, @pid
-    Process.wait @pid
+    wait_server
 
     assert_raises(Errno::ECONNREFUSED) { TCPSocket.new(HOST, @tcp_port) }
   end
@@ -136,20 +180,18 @@ class TestIntegrationSingle < TestIntegration
 
     log = File.read('t1-stdout')
 
-    File.unlink 't1-stdout' if File.file? 't1-stdout'
-    File.unlink 't1-pid' if File.file? 't1-pid'
-
     assert_match(%r!GET / HTTP/1\.1!, log)
+  ensure
+    File.unlink 't1-stdout' if File.file? 't1-stdout'
+    File.unlink 't1-pid'    if File.file? 't1-pid'
   end
 
   def test_puma_started_log_writing
     skip_unless_signal_exist? :TERM
 
-    suppress_output = '> /dev/null 2>&1'
-
     cli_server '-C test/config/t2_conf.rb test/rackup/hello.ru'
 
-    system "curl http://localhost:#{@tcp_port}/ #{suppress_output}"
+    system "curl http://localhost:#{@tcp_port}/ > /dev/null 2>&1"
 
     out=`#{BASE} bin/pumactl -F test/config/t2_conf.rb status`
 
@@ -157,26 +199,24 @@ class TestIntegrationSingle < TestIntegration
 
     log = File.read('t2-stdout')
 
-    File.unlink 't2-stdout' if File.file? 't2-stdout'
-
     assert_match(%r!GET / HTTP/1\.1!, log)
     assert(!File.file?("t2-pid"))
     assert_equal("Puma is started\n", out)
+  ensure
+    File.unlink 't2-stdout' if File.file? 't2-stdout'
   end
 
   def test_application_logs_are_flushed_on_write
-    @control_tcp_port = UniquePort.call
-    cli_server "--control-url tcp://#{HOST}:#{@control_tcp_port} --control-token #{TOKEN} test/rackup/write_to_stdout.ru"
+    cli_server "#{set_pumactl_args} test/rackup/write_to_stdout.ru"
 
     read_body connect
 
     cli_pumactl 'stop'
 
-    assert_equal "hello\n", @server.gets
-    assert_includes @server.read, 'Goodbye!'
+    assert wait_for_server_to_include("hello\n")
+    assert wait_for_server_to_include("Goodbye!")
 
-    @server.close unless @server.closed?
-    @server = nil
+    wait_server
   end
 
   # listener is closed 'externally' while Puma is in the IO.select statement
@@ -184,7 +224,12 @@ class TestIntegrationSingle < TestIntegration
     skip_unless_signal_exist? :TERM
 
     cli_server "test/rackup/close_listeners.ru", merge_err: true
-    read_body connect
+    connection = fast_connect
+
+    begin
+      read_body connection
+    rescue EOFError
+    end
 
     begin
       Timeout.timeout(5) do
@@ -203,4 +248,50 @@ class TestIntegrationSingle < TestIntegration
     end
     assert true
   end
+
+  def test_puma_debug_loaded_exts
+    cli_server "#{set_pumactl_args} test/rackup/hello.ru", puma_debug: true
+
+    assert wait_for_server_to_include('Loaded Extensions:')
+
+    cli_pumactl 'stop'
+    assert wait_for_server_to_include('Goodbye!')
+    wait_server
+  end
+
+  def test_idle_timeout
+    cli_server "test/rackup/hello.ru", config: "idle_timeout 1"
+
+    connect
+
+    sleep 1.15
+
+    assert_raises Errno::ECONNREFUSED, "Connection refused" do
+      connect
+    end
+  end
+
+  def test_pre_existing_unix_after_idle_timeout
+    skip_unless :unix
+
+    File.open(@bind_path, mode: 'wb') { |f| f.puts 'pre existing' }
+
+    cli_server "-q test/rackup/hello.ru", unix: :unix, config: "idle_timeout 1"
+
+    sock = connection = connect(nil, unix: true)
+    read_body(connection)
+
+    sleep 1.15
+
+    assert sock.wait_readable(1), 'Unexpected timeout'
+    assert_raises Puma.jruby? ? IOError : Errno::ECONNREFUSED, "Connection refused" do
+      connection = connect(nil, unix: true)
+    end
+
+    assert File.exist?(@bind_path)
+  ensure
+    if UNIX_SKT_EXIST
+      File.unlink @bind_path if File.exist? @bind_path
+    end
+  end
 end
