@@ -1,3 +1,5 @@
+# frozen_string_literal: true
+
 require_relative "helper"
 require_relative "helpers/test_puma/puma_socket"
 require "puma/events"
@@ -10,7 +12,7 @@ class WithoutBacktraceError < StandardError
   def message; "no backtrace error"; end
 end
 
-class TestPumaServer < TimeoutTestCase
+class TestPumaServer < PumaTest
   parallelize_me!
 
   include TestPuma
@@ -52,6 +54,7 @@ class TestPumaServer < TimeoutTestCase
     assert_equal "HTTP/1.0 200 OK", response.status
     assert_equal "HTTP/1.0"       , response.body
   end
+
   def test_http11_req_to_http11_resp
     server_run do |env|
       [200, {}, [env["SERVER_PROTOCOL"]]]
@@ -159,6 +162,51 @@ class TestPumaServer < TimeoutTestCase
     tf&.close
   end
 
+  def test_pipe_body_http11
+    skip if ::Puma::IS_WINDOWS && RUBY_VERSION < '2.7' #encoding issue?
+
+    random_bytes = SecureRandom.random_bytes(4096)
+
+    r, w = IO.pipe(binmode: true)
+
+    w.write random_bytes
+    w.close
+
+    server_run { |env| [200, {}, r] }
+
+    response = send_http_read_response
+    body = response.decode_body
+
+    assert_equal random_bytes.bytesize, body.bytesize
+    assert_equal random_bytes, body
+  ensure
+    w&.close
+    r&.close
+  end
+
+  def test_pipe_body_http10
+    skip if ::Puma::IS_WINDOWS && RUBY_VERSION < '2.7' #encoding issue?
+
+    bytes = 4_096
+    random_bytes = SecureRandom.random_bytes(bytes)
+
+    r, w = IO.pipe(binmode: true)
+
+    w.write random_bytes
+    w.close
+
+    server_run { |env| [200, {'content-length' => bytes.to_s}, r] }
+
+    response = send_http_read_response GET_10
+    body = response.body
+
+    assert_equal random_bytes.bytesize, body.bytesize
+    assert_equal random_bytes, body
+  ensure
+    w&.close
+    r&.close
+  end
+
   def test_proper_stringio_body
     data = nil
 
